#!/usr/bin/env python3
import os
import sys
import re
import argparse
import yaml
import cashaddress
from typing import Union

file_path = os.path.abspath(os.path.join(os.path.dirname(__file__)))
sys.path.insert(0, os.path.abspath(file_path))

import mnemonic_util
import wif_util
import p2pkh_util
import p2wpkh_util
import p2sh_p2wpkh_util
import p2tr_util
import common_util


def get_base58_prefix(coins_info, chain: str, typ: str) -> Union[bytes, None]:
    if coins_info[chain]["base58_prefix"][typ] is None:
        return None
    else:
        return coins_info[chain]["base58_prefix"][typ].to_bytes(1, byteorder='big')


def get_bech32_hrp(coins_info, chain: str) -> Union[str, None]:
    return coins_info[chain]["bech32_hrp"]


def get_derivation_path(coins_info, chain: str, typ: str) -> Union[str, None]:
    try:
        derivation_path = coins_info[chain]["hd_path"][typ]
    except KeyError:
        derivation_path = None
    return derivation_path


def main_entry(argv):
    mnemonic = ''
    private_key = b''
    private_key_wif = b''
    private_key_wif_compressed = b''
    public_key_uncompressed = b''
    public_key_compressed = b''
    public_key_uncompressed_hash160 = b''
    public_key_compressed_hash160 = b''
    public_key_hash160 = b''
    taproot_tweaked_private_key = b''
    taproot_tweaked_public_key = b''
    addr_p2pkh_uncompressed = b''
    addr_p2pkh_compressed = b''
    addr_p2pkh = b''  # uncompressed or compressed
    addr_p2sh_p2wpkh = b''
    addr_p2wpkh = ''
    addr_p2tr = ''

    file = open(os.path.join(os.path.abspath(file_path), "coins.yaml"), 'r', encoding="utf-8")
    file_data = file.read()
    file.close()

    coins_info = yaml.load(file_data, Loader=yaml.FullLoader)

    parser = argparse.ArgumentParser(description='A utility for dump btc address')
    parser.add_argument(
        "-c",
        "--chain",
        help="specify chain, default is btc",
        default="btc",
        dest="chain",
        choices=coins_info.keys())
    parser.add_argument(
        "-d",
        "--derivation",
        help="specify derivation path scheme (bip44/bip49/bip84/bip86) or customized path (e.g. m/0'/0/0), default is "
             "bip44. Only applicable when input mnemonic words",
        default="bip44",
        dest="derivation")
    parser.add_argument("inputs", metavar="mnemonic-words|private-key|public-key")
    args = parser.parse_args(argv[1:])
    chain = args.chain
    derivation = args.derivation
    inputs = args.inputs

    # See https://en.bitcoin.it/wiki/List_of_address_prefixes
    pubkey_version_bytes = get_base58_prefix(coins_info, chain, 'pubkey')  # 0x00 for btc mainnet, 0x6f for testnet
    script_version_bytes = get_base58_prefix(coins_info, chain, 'script')  # 0x05 for btc mainnet, 0xc4 for testnet
    wif_version_bytes = get_base58_prefix(coins_info, chain, 'wif')  # 0x80 for btc mainnet, 0xef for testnet
    human_readable_part = get_bech32_hrp(coins_info, chain)  # "bc" for btc mainnet, and "tb" for testnet

    if re.search("^([a-zA-Z]+\\s){11}([a-zA-Z]+).*$", inputs):
        # 12 mnemonic words
        # For example: olympic wine chicken argue unaware bundle tunnel grid spider slot spell need
        # sys.stderr.write("you input mnemonic\n")
        mnemonic = inputs
        derivation_path = derivation if derivation.startswith("m/") else get_derivation_path(coins_info, chain, derivation)
        if derivation_path is None or len(derivation_path) == 0:
            sys.stderr.write("derivation path {} is not available for {}\n".format(derivation, chain))
            sys.exit(1)
        private_key = mnemonic_util.mnemonic_to_private_key(mnemonic, derivation_path)
        if chain == "grs" or chain == "grs_testnet":
            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        else:
            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        public_key_compressed = common_util.prikey_to_pubkey(private_key, compressed=True)
        public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
        if derivation == "bip44":
            # For legacy address
            if chain == "grs" or chain == "grs_testnet":
                addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
            else:
                addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        elif derivation == "bip49":
            # For p2sh-segwit address
            if script_version_bytes:
                if chain == "grs" or chain == "grs_testnet":
                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
                else:
                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
        elif derivation == "bip84":
            # For bech32 address
            if human_readable_part:
                addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
        elif derivation == "bip86":
            # For bech32m address
            if human_readable_part:
                taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
                public_key_x_coordinate = public_key_compressed[1:33]
                taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
        elif derivation.startswith("m/"):  # customized path
            # For legacy address
            if chain == "grs" or chain == "grs_testnet":
                addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
            else:
                addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
            # For p2sh-segwit address
            if script_version_bytes:
                if chain == "grs" or chain == "grs_testnet":
                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
                else:
                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
            # For bech32 address
            if human_readable_part:
                addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
            # For bech32m address
            if human_readable_part:
                taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
                public_key_x_coordinate = public_key_compressed[1:33]
                taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
    elif (len(inputs) == 66 and inputs.startswith("0x")) or len(inputs) == 64:
        # sys.stderr.write("you input private key\n")
        # private key
        # For example: 0xc7ac679b56f50bfd54dd924fe45a8dca7a1c2dced254b03dac22afc03adb9127
        # For example: c7ac679b56f50bfd54dd924fe45a8dca7a1c2dced254b03dac22afc03adb9127
        private_key_hex = inputs.lower().replace('0x', '')
        private_key = bytes.fromhex(private_key_hex)
        if chain == "grs" or chain == "grs_testnet":
            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        else:
            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        public_key_uncompressed = common_util.prikey_to_pubkey(private_key, compressed=False)
        public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
        public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
        public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
        if chain == "grs" or chain == "grs_testnet":
            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        else:
            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        if script_version_bytes:
            if chain == "grs" or chain == "grs_testnet":
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
            else:
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
        if human_readable_part:
            addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
            taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
            public_key_x_coordinate = public_key_compressed[1:33]
            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
    elif (len(inputs) == 130 and inputs.startswith("0x")) or len(inputs) == 128 \
            or (len(inputs) == 132 and inputs.startswith("0x04")) \
            or (len(inputs) == 130 and inputs.startswith("04")):
        # sys.stderr.write("you input uncompressed public key\n")
        # public key
        # For example: 0x4cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1087190d91e26af594e3f8ecd3f4d3596c03c45d3b235da916903c930c6593cc4
        # For example: 4cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1087190d91e26af594e3f8ecd3f4d3596c03c45d3b235da916903c930c6593cc4
        # For example: 0x044cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1087190d91e26af594e3f8ecd3f4d3596c03c45d3b235da916903c930c6593cc4
        # For example: 044cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1087190d91e26af594e3f8ecd3f4d3596c03c45d3b235da916903c930c6593cc4
        public_key_hex = inputs[-128:]  # keep last 128 (remove leading 0x04, 0x, 04)
        public_key_uncompressed = b'\04' + bytes.fromhex(public_key_hex)
        public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
        public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
        public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
        if chain == "grs" or chain == "grs_testnet":
            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        else:
            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        if script_version_bytes:
            if chain == "grs" or chain == "grs_testnet":
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
            else:
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
        if human_readable_part:
            addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
            public_key_x_coordinate = public_key_compressed[1:33]
            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
    elif (len(inputs) == 68 and inputs.startswith("0x")) or len(inputs) == 66:
        # sys.stderr.write("you input compressed public key\n")
        # compressed public key
        # For example: 0x024cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1
        # For example: 024cd0aaeca3b636078583408e75edd77307b5190ca7a48bb9fbc1f2576c17dff1
        public_key_compressed_hexstr = inputs.lower().replace('0x', '')
        public_key_compressed = bytes.fromhex(public_key_compressed_hexstr)
        public_key_uncompressed = common_util.pubkey_compressed_to_uncompressed(public_key_compressed)
        public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
        public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
        if chain == "grs" or chain == "grs_testnet":
            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        else:
            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        if script_version_bytes:
            if chain == "grs" or chain == "grs_testnet":
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
            else:
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
        if human_readable_part:
            addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
            public_key_x_coordinate = public_key_compressed[1:33]
            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
    elif (len(inputs) == 42 and inputs.startswith("0x")) or len(inputs) == 40:
        # sys.stderr.write("you input hash160 of public key\n")
        public_key_hash160 = bytes.fromhex(inputs.lower().replace('0x', ''))
        if chain == "grs" or chain == "grs_testnet":
            addr_p2pkh = p2pkh_util.grs_hash160_to_p2pkh_addr(public_key_hash160, pubkey_version_bytes)
        else:
            addr_p2pkh = p2pkh_util.hash160_to_p2pkh_addr(public_key_hash160, pubkey_version_bytes)
        if script_version_bytes:
            if chain == "grs" or chain == "grs_testnet":
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_hash160_to_p2sh_p2wpkh_addr(public_key_hash160, script_version_bytes)
            else:
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.hash160_to_p2sh_p2wpkh_addr(public_key_hash160, script_version_bytes)
        if human_readable_part:
            addr_p2wpkh = p2wpkh_util.hash160_to_segwit_v0_addr(human_readable_part, public_key_hash160)
    elif wif_util.is_valid_wif(inputs, chain):
        if chain == "btc" or chain == "grs":
            assert inputs.startswith('5') or inputs.startswith('K') or inputs.startswith('L')
        if chain == "btc-test" or chain == "grs_testnet":
            assert inputs.startswith('9') or inputs.startswith('c')
        if chain == "grs" or chain == "grs_testnet":
            private_key = wif_util.grs_decode_wif(inputs)
        else:
            private_key = wif_util.decode_wif(inputs)
        if chain == "grs" or chain == "grs_testnet":
            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        else:
            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
        public_key_uncompressed = common_util.prikey_to_pubkey(private_key, compressed=False)
        public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
        public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
        public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
        if chain == "grs" or chain == "grs_testnet":
            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        else:
            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
        if script_version_bytes:
            if chain == "grs" or chain == "grs_testnet":
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
            else:
                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
        if human_readable_part:
            addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
            taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
            public_key_x_coordinate = public_key_compressed[1:33]
            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
    else:
        sys.stderr.write("invalid input: {0}\n".format(inputs))
        sys.exit(1)

    if mnemonic:
        print("mnemonic = {}".format(mnemonic))
    if private_key:
        print("private key (hex) = {}".format(private_key.hex()))
    if private_key_wif:
        print("private key (WIF) = {}".format(private_key_wif.decode('ascii')))
    if private_key_wif_compressed:
        print("private key (WIF compressed) = {}".format(private_key_wif_compressed.decode('ascii')))
    if taproot_tweaked_private_key:
        print("taproot tweaked private key = {}".format(taproot_tweaked_private_key.hex()))
    if public_key_uncompressed:
        print("public key (uncompressed) = {}".format(public_key_uncompressed.hex()))
    if public_key_compressed:
        print("public key (compressed) = {}".format(public_key_compressed.hex()))
    if public_key_uncompressed_hash160:
        print("hash160 of uncompressed public key = {}".format(public_key_uncompressed_hash160.hex()))
    if public_key_compressed_hash160:
        print("hash160 of compressed public key = {}".format(public_key_compressed_hash160.hex()))
    if public_key_hash160:
        print("hash160 of public key = {}".format(public_key_hash160.hex()))
    if taproot_tweaked_public_key:
        print("taproot tweaked public key (taproot output key) = {}".format(taproot_tweaked_public_key.hex()))
    if addr_p2pkh_uncompressed:
        print("legacy address (p2pkh uncompressed) = {}".format(addr_p2pkh_uncompressed.decode('ascii')))
        if chain == "bch":
            print("bitcoin cash address (p2pkh uncompressed) = {}".format(cashaddress.convert.to_cash_address(addr_p2pkh_uncompressed.decode('ascii'))))
    if addr_p2pkh_compressed:
        print("legacy address (p2pkh compressed) = {}".format(addr_p2pkh_compressed.decode('ascii')))
        if chain == "bch":
            print("bitcoin cash address (p2pkh compressed) = {}".format(cashaddress.convert.to_cash_address(addr_p2pkh_compressed.decode('ascii'))))
    if addr_p2pkh:
        print("legacy address (p2pkh) = {}".format(addr_p2pkh.decode('ascii')))
        if chain == "bch":
            print("bitcoin cash address (p2pkh) = {}".format(cashaddress.convert.to_cash_address(addr_p2pkh.decode('ascii'))))
    if addr_p2sh_p2wpkh:
        if public_key_hash160:
            print("p2sh-segwit address (only valid if input is hash160 of COMPRESSED public key) = {}".format(
                addr_p2sh_p2wpkh.decode('ascii')))
        else:
            print("p2sh-segwit address (p2sh p2wpkh) = {}".format(addr_p2sh_p2wpkh.decode('ascii')))
    if addr_p2wpkh:
        if public_key_hash160:
            print("bech32 address (only valid if input is hash160 of COMPRESSED public key) = {}".format(addr_p2wpkh))
        else:
            print("bech32 address (p2wpkh) = {}".format(addr_p2wpkh))
    if addr_p2tr:
        print("bech32m address (p2tr) = {}".format(addr_p2tr))


if __name__ == '__main__':
    main_entry(sys.argv)



import base58
import base58grs
import os
import sys

file_path = os.path.abspath(os.path.join(os.path.dirname(__file__)))
sys.path.insert(0, os.path.abspath(file_path))

import common_util


def pubkey_to_hash160(pubkey: bytes) -> bytes:
    """ Computes ripemd160(sha256(pubkey)) """
    out1 = common_util.sha256(pubkey)
    out2 = common_util.ripemd160(out1)
    return out2


def hash160_to_p2pkh_addr(hash160: bytes, version: bytes) -> bytes:
    """ Derives legacy (p2pkh) address from hash160 of pubkey """
    # Base-58 encoding with a checksum
    checksum = common_util.base58_cksum(version + hash160)
    address = base58.b58encode(version + hash160 + checksum)
    return address


def pubkey_to_p2pkh_addr(pubkey: bytes, version: bytes) -> bytes:
    """ Derives legacy (p2pkh) address from pubkey """
    out1 = common_util.sha256(pubkey)
    out2 = common_util.ripemd160(out1)
    # Base-58 encoding with a checksum
    checksum = common_util.base58_cksum(version + out2)
    address = base58.b58encode(version + out2 + checksum)
    return address


def grs_hash160_to_p2pkh_addr(hash160: bytes, version: bytes) -> bytes:
    """ Derives legacy (p2pkh) address from hash160 of pubkey """
    # Base-58 encoding with a checksum
    checksum = common_util.grs_base58_cksum(version + hash160)
    address = base58grs.b58encode(version + hash160 + checksum)
    return address


def grs_pubkey_to_p2pkh_addr(pubkey: bytes, version: bytes) -> bytes:
    """ Derives legacy (p2pkh) address from pubkey """
    out1 = common_util.sha256(pubkey)
    out2 = common_util.ripemd160(out1)
    # Base-58 encoding with a checksum
    checksum = common_util.grs_base58_cksum(version + out2)
    address = base58grs.b58encode(version + out2 + checksum)
    return address



import ecdsa
import hashlib
import groestlcoin_hash
from typing import Tuple
from ecdsa.ellipticcurve import PointJacobi


def get_x(point: PointJacobi) -> int:
    # Return point's x coordinate with type int
    # If gmpy/gmpy2 is installed, point.x() return type of mpz (see https://gmpy2.readthedocs.io/en/latest/mpz.html)
    # If gmpy/gmpy2 is not installed, point.x() return type of int
    return int(point.x())


def get_y(point: PointJacobi) -> int:
    # Return point's y coordinate with type int
    return int(point.y())


def sha256(inputs: bytes) -> bytes:
    """ Computes sha256 """
    sha = hashlib.sha256()
    sha.update(inputs)
    return sha.digest()


def ripemd160(inputs: bytes) -> bytes:
    """ Computes ripemd160 """
    rip = hashlib.new('ripemd160')
    rip.update(inputs)
    return rip.digest()


def base58_cksum(inputs: bytes) -> bytes:
    """ Computes base 58 four bytes check sum """
    s1 = sha256(inputs)
    s2 = sha256(s1)
    checksum = s2[0:4]
    return checksum
    

def grs_base58_cksum(inputs: bytes) -> bytes:
    """ Computes base 58 four bytes check sum for grs """
    return groestlcoin_hash.getHash(inputs, len(inputs))[0:4]


def pubkey_compressed_to_uncompressed(compressed_pubkey: bytes) -> bytes:
    """ Converts compressed pubkey to uncompressed format """
    assert len(compressed_pubkey) == 33
    # modulo p which is defined by secp256k1's spec
    p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
    x = int.from_bytes(compressed_pubkey[1:33], byteorder='big')
    y_sq = (pow(x, 3, p) + 7) % p
    y = pow(y_sq, (p + 1) // 4, p)
    if compressed_pubkey[0] % 2 != y % 2:
        y = p - y
    y_bytes = y.to_bytes(32, byteorder='big')
    return b'\04' + compressed_pubkey[1:33] + y_bytes  # x + y


def pubkey_uncompressed_to_compressed(uncompressed_pubkey: bytes) -> bytes:
    """ Converts uncompressed pubkey to compressed format """
    assert len(uncompressed_pubkey) == 65
    # Compressed public key is:
    # 0x02 + x - coordinate if y is even
    # 0x03 + x - coordinate if y is odd
    x = int.from_bytes(uncompressed_pubkey[1:33], byteorder='big')  # uncompressed_pubkey must contain prefix b'\04'
    y = int.from_bytes(uncompressed_pubkey[33:65], byteorder='big')
    parity = y & 1
    compressed_public_key = (2 + parity).to_bytes(1, byteorder='big') + x.to_bytes(32, byteorder='big')
    return compressed_public_key


def pubkey_from_bytes_to_point(pubkey: bytes) -> Tuple[int, int]:
    """ Returns the x y coordinate of a pubkey (can be uncompressed or compressed) """
    assert len(pubkey) == 33 or len(pubkey) == 65
    if len(pubkey) == 33:  # compressed pubkey
        uncompressed_pubkey = pubkey_compressed_to_uncompressed(pubkey)
    else:
        uncompressed_pubkey = pubkey
    x = int.from_bytes(uncompressed_pubkey[1:33], byteorder='big')  # uncompressed_pubkey must contain prefix b'\04'
    y = int.from_bytes(uncompressed_pubkey[33:65], byteorder='big')
    return x, y


def pubkey_from_point_to_bytes(x: int, y: int, compressed: bool = True) -> bytes:
    """ Constructs pubkey from its x y coordinates """
    x_bytes = x.to_bytes(32, byteorder='big')
    y_bytes = y.to_bytes(32, byteorder='big')
    if compressed:
        parity = y & 1
        return (2 + parity).to_bytes(1, byteorder='big') + x_bytes
    else:
        return b'\04' + x_bytes + y_bytes


def prikey_to_pubkey(private_key: bytes, compressed: bool = True) -> bytes:
    """ Derives pubkey from private key """
    pubkey_point: PointJacobi = int.from_bytes(private_key, byteorder='big') * ecdsa.curves.SECP256k1.generator
    return pubkey_from_point_to_bytes(get_x(pubkey_point), get_y(pubkey_point), compressed)



import hashlib
import os
import sys
from typing import Optional

import ecdsa
from ecdsa.ellipticcurve import PointJacobi, Point

file_path = os.path.abspath(os.path.join(os.path.dirname(__file__)))
sys.path.insert(0, os.path.abspath(file_path))

import common_util


def bytes_from_int(x: int) -> bytes:
    return x.to_bytes(32, byteorder="big")


def int_from_bytes(b: bytes) -> int:
    return int.from_bytes(b, byteorder="big")


def tagged_hash(tag: str, msg: bytes) -> bytes:
    tag_hash = hashlib.sha256(tag.encode()).digest()
    return hashlib.sha256(tag_hash + tag_hash + msg).digest()


def lift_x(x: bytes) -> Optional[PointJacobi]:
    """
    Define in BIP 340, see https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki

    :param x: an integer in range 0..p-1,
    :return: the point P for which P.x() = x and has_even_y(P), or None if no such point exists.
    """
    p = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F
    x = int_from_bytes(x)
    if x >= p:
        return None
    y_sq = (pow(x, 3, p) + 7) % p
    y = pow(y_sq, (p + 1) // 4, p)
    if pow(y, 2, p) != y_sq:
        return None
    P: PointJacobi = PointJacobi.from_affine(Point(ecdsa.curves.SECP256k1.curve, x, y if y % 2 == 0 else p - y))
    return P


def has_even_y(P: PointJacobi) -> bool:
    return common_util.get_y(P) % 2 == 0


def taproot_tweak_seckey(seckey0: bytes) -> bytes:
    """
    See https://github.com/bitcoin/bips/blob/master/bip-0341.mediawiki
    :param seckey0: the private key
    :return: tweaked private key
    """
    P = int_from_bytes(seckey0) * ecdsa.curves.SECP256k1.generator
    seckey = int_from_bytes(seckey0) if has_even_y(P) else ecdsa.curves.SECP256k1.order - int_from_bytes(seckey0)
    t = int_from_bytes(tagged_hash("TapTweak", bytes_from_int(common_util.get_x(P))))
    if t >= ecdsa.curves.SECP256k1.order:
        raise ValueError
    return bytes_from_int(int((seckey + t) % ecdsa.curves.SECP256k1.order))


def public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate: bytes) -> bytes:
    """
    See https://github.com/bitcoin/bips/blob/master/bip-0086.mediawiki

    P:                  lift_x(public_key_x_coordinate)
    internal_key:       P.x()
    Q:                  P + int(HashTapTweak(bytes(P.x()))) * G
    32_byte_output_key: bytes(Q.x())
    """
    P: PointJacobi = lift_x(public_key_x_coordinate)
    assert P is not None
    t = int_from_bytes(tagged_hash("TapTweak", bytes_from_int(common_util.get_x(P))))
    if t >= ecdsa.curves.SECP256k1.order:
        raise ValueError
    Q: PointJacobi = P + t * ecdsa.curves.SECP256k1.generator
    return bytes_from_int(common_util.get_x(Q))



import hashlib
import groestlcoin_hash
import base58
import base58grs
from typing import Union


def scrub_input(hex_str_or_bytes: Union[str, bytes]) -> bytes:
    if isinstance(hex_str_or_bytes, str):
        hex_str_or_bytes = bytes.fromhex(hex_str_or_bytes)
    return hex_str_or_bytes


# wallet import format key - base58 encoded format
# https://bitcoin.stackexchange.com/questions/9244/private-key-to-wif
def encode_wif(private_key: Union[str, bytes], version: bytes, compressed_wif: bool = False) -> bytes:
    """ Encode wif format private key """
    private_key = scrub_input(private_key)

    # prepended version byte to private key
    private_key_with_version = version + private_key
    if compressed_wif:
        private_key_with_version = version + private_key + b'\x01'
    # perform SHA-256 hash on the mainnet_private_key
    sha256 = hashlib.sha256()
    sha256.update(private_key_with_version)
    hash_bytes = sha256.digest()

    # perform SHA-256 on the previous SHA-256 hash
    sha256 = hashlib.sha256()
    sha256.update(hash_bytes)
    hash_bytes = sha256.digest()

    # create a checksum using the first 4 bytes of the previous SHA-256 hash
    # append the 4 checksum bytes to the mainnet_private_key
    checksum = hash_bytes[:4]

    # print('checksum', checksum.hex())
    hash_bytes = private_key_with_version + checksum
    # print('hash', hash_bytes.hex())

    # convert private_key_with_version + checksum into base58 encoded string
    return base58.b58encode(hash_bytes)


def grs_encode_wif(private_key: Union[str, bytes], version: bytes, compressed_wif: bool = False) -> bytes:
    """ Encode wif format private key """
    private_key = scrub_input(private_key)

    # prepended version byte to private key
    private_key_with_version = version + private_key
    if compressed_wif:
        private_key_with_version = version + private_key + b'\x01'
    # perform DOUBLE GROESTL-512 hash on the mainnet_private_key
    hash_bytes = groestlcoin_hash.getHash(private_key_with_version, len(private_key_with_version))

    # create a checksum using the first 4 bytes of the GROESTL-512 hash
    # append the 4 checksum bytes to the mainnet_private_key
    checksum = hash_bytes[:4]

    # print('checksum', checksum.hex())
    hash_bytes = private_key_with_version + checksum
    # print('hash', hash_bytes.hex())

    # convert private_key_with_version + checksum into base58 encoded string
    return base58grs.b58encode(hash_bytes)


def decode_wif(wif: str, compressed_wif: bool = False) -> bytes:
    """ Decode wif format private key """
    decoded = base58.b58decode(wif)
    if len(decoded) == 38 or len(decoded) == 37:
        private_key = decoded[1:33]  # WIF-compressed: [0x80 xxxx 0x01 checksum] or WIF: [0x80 xxxx checksum]
    else:
        raise ValueError("invalid WIF")
    return private_key


def grs_decode_wif(wif: str, compressed_wif: bool = False) -> bytes:
    """ Decode wif format private key """
    decoded = base58grs.b58decode(wif)
    if len(decoded) == 38 or len(decoded) == 37:
        private_key = decoded[1:33]  # WIF-compressed: [0x80 xxxx 0x01 checksum] or WIF: [0x80 xxxx checksum]
    else:
        raise ValueError("invalid WIF")
    return private_key


def is_valid_wif(wif: str, chain: str) -> bool:
    """ Check if wif is valid"""
    try:
        if chain == "grs" or chain == "grs_testnet":
            result = base58grs.b58decode(wif)
            #print(result)
            result, check = result[:-4], result[-4:]
            # DOUBLE GROESTL-512 hash
            digest = groestlcoin_hash.getHash(result, len(result))
        else:
            result = base58.b58decode(wif)
            #print(result)
            result, check = result[:-4], result[-4:]
            # DOUBLE sha256 hash
            digest = hashlib.sha256(hashlib.sha256(result).digest()).digest()
        return check == digest[:4]
    except ValueError:
        return False
